@@ -1,3 +1,6 @@
+#![warn(clippy::pedantic, clippy::nursery)]
+#![allow(clippy::non_ascii_literal)]
+
 #[cfg(test)]
 mod tests;
 use unicode_segmentation::UnicodeSegmentation;
@@ -5,10 +8,10 @@ use unicode_segmentation::UnicodeSegmentation;
 #[derive(Copy, Clone, PartialEq, Eq, Debug, Hash)]
 #[allow(clippy::struct_excessive_bools)]
 pub struct PinyinParser {
-    _strict: bool,
-    _preserve_punctuations: bool,
-    _preserve_spaces: bool,
-    _preserve_miscellaneous: bool,
+    p_strict: bool,
+    p_preserve_punctuations: bool,
+    p_preserve_spaces: bool,
+    p_preserve_miscellaneous: bool,
 }
 
 impl Default for PinyinParser {
@@ -18,46 +21,82 @@ impl Default for PinyinParser {
 }
 
 impl PinyinParser {
-    pub fn new() -> Self {
-        PinyinParser {
-            _strict: false,
-            _preserve_spaces: false,
-            _preserve_punctuations: false,
-            _preserve_miscellaneous: false,
+    #[must_use]
+    pub const fn new() -> Self {
+        Self {
+            p_strict: false,
+            p_preserve_spaces: false,
+            p_preserve_punctuations: false,
+            p_preserve_miscellaneous: false,
         }
     }
 
-    pub fn is_strict(self, b: bool) -> Self {
-        Self { _strict: b, ..self }
-    }
-
-    pub fn preserve_spaces(self, b: bool) -> Self {
+    #[must_use]
+    pub const fn is_strict(self, b: bool) -> Self {
         Self {
-            _preserve_spaces: b,
+            p_strict: b,
             ..self
         }
     }
 
-    pub fn preserve_punctuations(self, b: bool) -> Self {
+    #[must_use]
+    pub const fn preserve_spaces(self, b: bool) -> Self {
         Self {
-            _preserve_punctuations: b,
+            p_preserve_spaces: b,
             ..self
         }
     }
 
-    pub fn preserve_miscellaneous(self, b: bool) -> Self {
+    #[must_use]
+    pub const fn preserve_punctuations(self, b: bool) -> Self {
         Self {
-            _preserve_miscellaneous: b,
+            p_preserve_punctuations: b,
             ..self
         }
     }
 
+    /// ```
+    /// use pinyin_parser::PinyinParser;
+    /// let parser = PinyinParser::new()
+    ///     .is_strict(true)
+    ///     .preserve_miscellaneous(true);
+    /// assert_eq!(
+    ///     parser
+    ///         .parse("你Nǐ 好hǎo")
+    ///         .into_iter()
+    ///         .collect::<Vec<_>>(),
+    ///     vec!["你", "nǐ", "好", "hǎo"]
+    /// )
+    /// ```
+    #[must_use]
+    pub const fn preserve_miscellaneous(self, b: bool) -> Self {
+        Self {
+            p_preserve_miscellaneous: b,
+            ..self
+        }
+    }
+
+    /// ```
+    /// use pinyin_parser::PinyinParser;
+    /// let parser = PinyinParser::new()
+    ///     .is_strict(true)
+    ///     .preserve_punctuations(true)
+    ///     .preserve_spaces(true);
+    /// assert_eq!(
+    ///     parser
+    ///         .parse("Nǐ zuò shénme?")
+    ///         .into_iter()
+    ///         .collect::<Vec<_>>(),
+    ///     vec!["nǐ", " ", "zuò", " ", "shén", "me", "?"]
+    /// )
+    /// ```
+    #[must_use]
     pub fn parse(self, s: &str) -> PinyinParserIter {
         PinyinParserIter {
             configs: self,
             it: VecAndIndex {
                 vec: UnicodeSegmentation::graphemes(s, true)
-                    .map(|c| pinyin_token::to_token(c))
+                    .map(|c| pinyin_token::to_token(c, self.p_strict))
                     .collect::<Vec<_>>(),
                 next_pos: 0,
             },
@@ -65,10 +104,55 @@ impl PinyinParser {
         }
     }
 
+    /// Strict mode: 
+    /// * forbids the use of breve instead of hacek to represent the third tone
+    /// * forbids the use of IPA `ɡ` (U+0261) instead of `g`
+    /// ```
+    /// use pinyin_parser::PinyinParser;
+    /// assert_eq!(
+    ///     PinyinParser::strict("jīntiān")
+    ///         .into_iter()
+    ///         .collect::<Vec<_>>(),
+    ///     vec!["jīn", "tiān"]
+    /// );
+    /// ```
+    
+    /// ```should_panic
+    /// use pinyin_parser::PinyinParser;
+    /// assert_eq!(
+    ///     PinyinParser::strict("zǒnɡshì") // this `ɡ` is not the `g` from ASCII
+    ///         .into_iter()
+    ///         .collect::<Vec<_>>(),
+    ///     vec!["zǒng", "shì"]
+    /// );
+    /// ```
+
+    /// This parser supports the use of `ẑ`, `ĉ`, `ŝ` and `ŋ`, though I have never seen anyone use it.
+    /// ```
+    /// use pinyin_parser::PinyinParser;
+    /// assert_eq!(
+    ///     PinyinParser::strict("Ẑāŋ").into_iter().collect::<Vec<_>>(),
+    ///     vec!["zhāng"]
+    /// )
+    /// ```
+
+    #[must_use]
     pub fn strict(s: &str) -> PinyinParserIter {
         Self::new().is_strict(true).parse(s)
     }
 
+    /// ```
+    /// use pinyin_parser::PinyinParser;
+    /// assert_eq!(
+    ///     // 'ă' is LATIN SMALL LETTER A WITH BREVE and is not accepted in strict mode.  
+    ///     // The correct alphabet to use is 'ǎ'.  
+    ///     PinyinParser::loose("mián'ăo") 
+    ///         .into_iter()
+    ///         .collect::<Vec<_>>(),
+    ///     vec!["mián", "ǎo"]
+    /// );
+    /// ```
+    #[must_use]
     pub fn loose(s: &str) -> PinyinParserIter {
         Self::new().parse(s)
     }
@@ -120,10 +204,17 @@ impl<T> VecAndIndex<T> {
 
 impl Iterator for PinyinParserIter {
     type Item = String;
+
+    #[allow(clippy::too_many_lines)]
+    #[allow(clippy::cognitive_complexity)]
     fn next(&mut self) -> Option<Self::Item> {
         use pinyin_token::Alphabet;
-        use pinyin_token::PinyinToken::*;
-        use ParserState::*;
+        use pinyin_token::PinyinToken::{
+            Alph, Apostrophe, LightToneMarker, Others, Punctuation, Space,
+        };
+        use ParserState::{
+            AfterSyllablePossiblyConsumingApostrophe, BeforeWordInitial, InitialParsed, ZCSParsed,
+        };
         loop {
             match (self.it.next(), self.state) {
                 (
@@ -139,8 +230,9 @@ impl Iterator for PinyinParserIter {
                     Some(Apostrophe),
                     AfterSyllablePossiblyConsumingApostrophe | BeforeWordInitial,
                 ) => panic!("unexpected apostrophe found at the beginning of a word"),
-                (None, AfterSyllablePossiblyConsumingApostrophe) => return None,
-                (None, BeforeWordInitial) => return None,
+                (None, AfterSyllablePossiblyConsumingApostrophe | BeforeWordInitial) => {
+                    return None
+                }
                 (None, InitialParsed(initial)) => {
                     panic!("unexpected end of string found after {:?}", initial)
                 }
@@ -149,42 +241,40 @@ impl Iterator for PinyinParserIter {
                     Some(Punctuation(s)),
                     BeforeWordInitial | AfterSyllablePossiblyConsumingApostrophe,
                 ) => {
-                    if self.configs._preserve_punctuations {
+                    if self.configs.p_preserve_punctuations {
                         self.state = BeforeWordInitial;
-                        return Some((*s).to_owned());
-                    } else {
-                        continue;
+                        return Some((*s).clone());
                     }
+                    continue;
                 }
                 (Some(Space(s)), BeforeWordInitial | AfterSyllablePossiblyConsumingApostrophe) => {
-                    if self.configs._preserve_spaces {
+                    if self.configs.p_preserve_spaces {
                         self.state = BeforeWordInitial;
-                        return Some((*s).to_owned());
-                    } else {
-                        continue;
+                        return Some((*s).clone());
                     }
+                    continue;
                 }
 
                 (Some(Others(s)), BeforeWordInitial | AfterSyllablePossiblyConsumingApostrophe) => {
-                    if self.configs._preserve_miscellaneous {
+                    if self.configs.p_preserve_miscellaneous {
                         self.state = BeforeWordInitial;
-                        return Some((*s).to_owned());
-                    } else {
-                        continue;
+                        return Some((*s).clone());
                     }
+                    continue;
                 }
 
                 (
                     Some(Alph(alph)),
                     BeforeWordInitial | AfterSyllablePossiblyConsumingApostrophe,
-                ) => match alph.alphabet {
+                ) => {
+                    match alph.alphabet {
                     Alphabet::B => self.state = InitialParsed(SpellingInitial::B),
                     Alphabet::P => self.state = InitialParsed(SpellingInitial::P),
                     Alphabet::M => {
                         if alph.diacritics.is_empty() {
                             self.state = InitialParsed(SpellingInitial::M);
                         } else {
-                            return Some(alph.to_str(self.configs._strict));
+                            return Some(alph.to_str(self.configs.p_strict));
                         }
                     }
                     Alphabet::F => self.state = InitialParsed(SpellingInitial::F),
@@ -194,7 +284,7 @@ impl Iterator for PinyinParserIter {
                         if alph.diacritics.is_empty() {
                             self.state = InitialParsed(SpellingInitial::N)
                         } else {
-                            return Some(alph.to_str(self.configs._strict));
+                            return Some(alph.to_str(self.configs.p_strict));
                         }
                     }
                     Alphabet::L => self.state = InitialParsed(SpellingInitial::L),
@@ -216,7 +306,7 @@ impl Iterator for PinyinParserIter {
                         ) {
                             self.state = InitialParsed(SpellingInitial::ZH)
                         } else {
-                            return Some(alph.to_str(self.configs._strict));
+                            return Some(alph.to_str(self.configs.p_strict));
                         }
                     }
                     Alphabet::C => {
@@ -228,7 +318,7 @@ impl Iterator for PinyinParserIter {
                         ) {
                             self.state = InitialParsed(SpellingInitial::CH)
                         } else {
-                            return Some(alph.to_str(self.configs._strict));
+                            return Some(alph.to_str(self.configs.p_strict));
                         }
                     }
                     Alphabet::S => {
@@ -240,7 +330,7 @@ impl Iterator for PinyinParserIter {
                         ) {
                             self.state = InitialParsed(SpellingInitial::SH)
                         } else {
-                            return Some(alph.to_str(self.configs._strict));
+                            return Some(alph.to_str(self.configs.p_strict));
                         }
                     }
                     Alphabet::A | Alphabet::E | Alphabet::O => {
@@ -252,17 +342,16 @@ impl Iterator for PinyinParserIter {
                         "unexpected alphabet {:?} found at the beginning of a word",
                         alph.alphabet,
                     ),
-                },
+                }},
 
-                (Some(Alph(alph)), ZCSParsed(zcs)) => match alph.alphabet {
-                    Alphabet::H => {
+                (Some(Alph(alph)), ZCSParsed(zcs)) => {
+                    if alph.alphabet == Alphabet::H {
                         self.state = match zcs {
                             ZCS::Z => InitialParsed(SpellingInitial::ZH),
                             ZCS::C => InitialParsed(SpellingInitial::CH),
                             ZCS::S => InitialParsed(SpellingInitial::SH),
                         }
-                    }
-                    _ => {
+                    } else {
                         self.it.rewind(1);
                         self.state = match zcs {
                             ZCS::Z => InitialParsed(SpellingInitial::Z),
@@ -270,16 +359,16 @@ impl Iterator for PinyinParserIter {
                             ZCS::S => InitialParsed(SpellingInitial::S),
                         }
                     }
-                },
+                }
 
                 (Some(Alph(_)), InitialParsed(initial)) => {
-                    use finals::*;
+                    use finals::Candidate;
                     self.it.rewind(1);
-                    let candidates = self.it.get_candidates_without_rhotic(self.configs._strict);
+                    let candidates = self.it.get_candidates_without_rhotic(self.configs.p_strict);
 
                     if candidates.is_empty() {
                         panic!(
-                            "no adequate candidate is found, after the initial {:?}",
+                            "no adequate candidate for finals (-an, -ian, ...) is found, after the initial {:?}",
                             initial
                         );
                     }
@@ -345,16 +434,15 @@ impl Iterator for PinyinParserIter {
                                             initial,
                                             finals::FinalWithTone { fin, tone }
                                         ));
-                                    } else {
-                                        // this is rhotic
-                                        self.it.advance(1);
-                                        self.state = AfterSyllablePossiblyConsumingApostrophe;
-                                        return Some(format!(
-                                            "{}{}r",
-                                            initial,
-                                            finals::FinalWithTone { fin, tone }
-                                        ));
                                     }
+                                    // this is rhotic
+                                    self.it.advance(1);
+                                    self.state = AfterSyllablePossiblyConsumingApostrophe;
+                                    return Some(format!(
+                                        "{}{}r",
+                                        initial,
+                                        finals::FinalWithTone { fin, tone }
+                                    ));
                                 }
 
                                 Alphabet::G =>
@@ -381,11 +469,10 @@ impl Iterator for PinyinParserIter {
                                             initial,
                                             finals::FinalWithTone { fin, tone }
                                         ));
-                                    } else {
-                                        // this candidate is wrong
-                                        self.it.rewind(fin_len);
-                                        continue;
                                     }
+                                    // this candidate is wrong
+                                    self.it.rewind(fin_len);
+                                    continue;
                                 }
 
                                 Alphabet::N => {
@@ -409,11 +496,10 @@ impl Iterator for PinyinParserIter {
                                             initial,
                                             finals::FinalWithTone { fin, tone }
                                         ));
-                                    } else {
-                                        // this candidate is not good
-                                        self.it.rewind(fin_len);
-                                        continue;
                                     }
+                                    // this candidate is not good
+                                    self.it.rewind(fin_len);
+                                    continue;
                                 }
 
                                 _ => {
@@ -428,7 +514,7 @@ impl Iterator for PinyinParserIter {
                         }
                     }
                     panic!(
-                        "no adequate candidate found among candidates {:?}",
+                        "no adequate candidate for finals (-an, -ian, ...) found, among possible candidates {:?}",
                         candidates
                     )
                 }
@@ -440,14 +526,14 @@ impl Iterator for PinyinParserIter {
 mod finals;
 
 #[derive(Debug, Copy, Clone, PartialEq, Eq, Hash)]
-pub enum ZCS {
+enum ZCS {
     Z,
     C,
     S,
 }
 
 #[derive(Debug, Copy, Clone, PartialEq, Eq, Hash)]
-pub enum SpellingInitial {
+enum SpellingInitial {
     B,
     P,
     M,
@@ -477,7 +563,6 @@ pub enum SpellingInitial {
 impl std::fmt::Display for SpellingInitial {
     fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         match self {
-            /* FIXME: capitalization is not preserved */
             SpellingInitial::B => write!(f, "b"),
             SpellingInitial::P => write!(f, "p"),
             SpellingInitial::M => write!(f, "m"),
@@ -505,71 +590,3 @@ impl std::fmt::Display for SpellingInitial {
         }
     }
 }
-
-#[derive(Copy, Clone, PartialEq, Eq, Debug, Hash)]
-pub struct PinyinAmbiguousParser {
-    _preserve_punctuations: bool,
-    _preserve_spaces: bool,
-    _preserve_capitalization: bool,
-}
-
-impl Default for PinyinAmbiguousParser {
-    fn default() -> Self {
-        Self::new()
-    }
-}
-
-impl PinyinAmbiguousParser {
-    pub fn new() -> Self {
-        PinyinAmbiguousParser {
-            _preserve_spaces: false,
-            _preserve_capitalization: false,
-            _preserve_punctuations: false,
-        }
-    }
-
-    pub fn preserve_spaces(self, b: bool) -> Self {
-        Self {
-            _preserve_spaces: b,
-            ..self
-        }
-    }
-
-    pub fn preserve_capitalization(self, b: bool) -> Self {
-        Self {
-            _preserve_capitalization: b,
-            ..self
-        }
-    }
-
-    /// allow british spelling
-    pub fn preserve_capitalisation(self, b: bool) -> Self {
-        self.preserve_capitalization(b)
-    }
-
-    pub fn preserve_punctuations(self, b: bool) -> Self {
-        Self {
-            _preserve_spaces: b,
-            ..self
-        }
-    }
-
-    pub fn parse(self, s: &str) -> PinyinAmbiguousParserIter {
-        PinyinAmbiguousParserIter {
-            configs: self,
-            it: VecAndIndex {
-                vec: UnicodeSegmentation::graphemes(s, true)
-                    .map(|c| pinyin_token::to_token(c))
-                    .collect::<Vec<_>>(),
-                next_pos: 0,
-            },
-            state: ParserState::BeforeWordInitial,
-        }
-    }
-}
-
-pub struct PinyinAmbiguousParserIter {
-    configs: PinyinAmbiguousParser,
-    it: VecAndIndex<pinyin_token::PinyinToken>,
-    state: ParserState,
-}
